@@ -3,7 +3,7 @@ use warnings;
 
 package JMAP::Tester;
 # ABSTRACT: a JMAP client made for testing JMAP servers
-$JMAP::Tester::VERSION = '0.024';
+$JMAP::Tester::VERSION = '0.025';
 use Moo;
 
 use Crypt::Misc qw(decode_b64u encode_b64u);
@@ -419,9 +419,15 @@ has _logger => (
 
 #pod =method upload
 #pod
-#pod   my $result = $tester->upload($mime_type, $blob_ref, \%arg);
+#pod   my $result = $tester->upload(\%arg);
 #pod
-#pod This uploads the given blob, which should be given as a reference to a string.
+#pod Required arguments are:
+#pod
+#pod   accountId - the account for which we're uploading (no default)
+#pod   type      - the content-type we want to provide to the server
+#pod   blob      - the data to upload. Must be a reference to a string
+#pod
+#pod This uploads the given blob.
 #pod
 #pod The return value will either be a L<failure
 #pod object|JMAP::Tester::Result::Failure> or an L<upload
@@ -430,19 +436,32 @@ has _logger => (
 #pod =cut
 
 sub upload {
-  my ($self, $mime_type, $blob_ref) = @_;
+  my ($self, $arg) = @_;
   # TODO: support blob as handle or sub -- rjbs, 2016-11-17
 
+  my $uri = $self->upload_uri;
+
   Carp::confess("can't upload without upload_uri")
-    unless $self->upload_uri;
+    unless $uri;
+
+  for my $param (qw(accountId type blob)) {
+    my $value = $arg->{ $param };
+
+    Carp::confess("missing required parameter $param")
+      unless defined $value;
+
+    if ($param eq 'accountId') {
+      $uri =~ s/\{$param\}/$value/g;
+    }
+  }
 
   my $post = HTTP::Request->new(
-    POST => $self->upload_uri,
+    POST => $uri,
     [
-      'Content-Type' => $mime_type,
+      'Content-Type' => $arg->{type},
       $self->_maybe_auth_header,
     ],
-    $$blob_ref,
+    ${ $arg->{blob} },
   );
 
   # Or our sub below leaks us
@@ -860,7 +879,7 @@ JMAP::Tester - a JMAP client made for testing JMAP servers
 
 =head1 VERSION
 
-version 0.024
+version 0.025
 
 =head1 OVERVIEW
 
@@ -987,9 +1006,15 @@ C<munge_method_triple>, which can tweak the method however it likes.
 
 =head2 upload
 
-  my $result = $tester->upload($mime_type, $blob_ref, \%arg);
+  my $result = $tester->upload(\%arg);
 
-This uploads the given blob, which should be given as a reference to a string.
+Required arguments are:
+
+  accountId - the account for which we're uploading (no default)
+  type      - the content-type we want to provide to the server
+  blob      - the data to upload. Must be a reference to a string
+
+This uploads the given blob.
 
 The return value will either be a L<failure
 object|JMAP::Tester::Result::Failure> or an L<upload
